@@ -40,16 +40,20 @@
 #include "../shared/visualtestutil.h"
 #include "../shared/qtest_quickcontrols.h"
 
+#include <QtCore/qoperatingsystemversion.h>
 #include <QtGui/qpa/qwindowsysteminterface.h>
 #include <QtQuick/qquickview.h>
+#include <QtQuick/private/qquickpalette_p.h>
 #include <QtQuickTemplates2/private/qquickapplicationwindow_p.h>
 #include <QtQuickTemplates2/private/qquickcombobox_p.h>
 #include <QtQuickTemplates2/private/qquickdialog_p.h>
 #include <QtQuickTemplates2/private/qquickoverlay_p.h>
 #include <QtQuickTemplates2/private/qquickpopup_p.h>
+#include <QtQuickTemplates2/private/qquickpopupitem_p_p.h>
 #include <QtQuickTemplates2/private/qquickbutton_p.h>
 #include <QtQuickTemplates2/private/qquickslider_p.h>
 #include <QtQuickTemplates2/private/qquickstackview_p.h>
+#include <QtQuickTemplates2/private/qquickpopup_p_p.h>
 
 using namespace QQuickVisualTestUtil;
 
@@ -93,7 +97,10 @@ private slots:
     void countChanged();
     void toolTipCrashOnClose();
     void setOverlayParentToNull();
+    void tabFence();
     void invisibleToolTipOpen();
+    void centerInOverlayWithinStackViewItem();
+    void destroyDuringExitTransition();
 };
 
 void tst_QQuickPopup::initTestCase()
@@ -113,6 +120,7 @@ void tst_QQuickPopup::visible()
 {
     QFETCH(QString, source);
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -146,6 +154,7 @@ void tst_QQuickPopup::visible()
 void tst_QQuickPopup::state()
 {
     QQuickApplicationHelper helper(this, "applicationwindow.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -207,6 +216,7 @@ void tst_QQuickPopup::overlay()
     QFETCH(bool, dim);
 
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -295,18 +305,7 @@ void tst_QQuickPopup::overlay()
     QVERIFY(popup->isVisible());
     QVERIFY(overlay->isVisible());
 
-    struct TouchDeviceDeleter
-    {
-        static inline void cleanup(QTouchDevice *device)
-        {
-            QWindowSystemInterface::unregisterTouchDevice(device);
-            delete device;
-        }
-    };
-
-    QScopedPointer<QTouchDevice, TouchDeviceDeleter> device(new QTouchDevice);
-    device->setType(QTouchDevice::TouchScreen);
-    QWindowSystemInterface::registerTouchDevice(device.data());
+    QScopedPointer<QPointingDevice> device(QTest::createTouchDevice());
 
     QTest::touchEvent(window, device.data()).press(0, QPoint(1, 1));
     QCOMPARE(overlayPressedSignal.count(), ++overlayPressCount);
@@ -369,6 +368,7 @@ void tst_QQuickPopup::zOrder()
 {
     QFETCH(QString, source);
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -478,6 +478,7 @@ void tst_QQuickPopup::closePolicy()
     QFETCH(QQuickPopup::ClosePolicy, closePolicy);
 
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -562,6 +563,7 @@ void tst_QQuickPopup::activeFocusOnClose1()
     // Test that a popup that never sets focus: true (e.g. ToolTip) doesn't affect
     // the active focus item when it closes.
     QQuickApplicationHelper helper(this, QStringLiteral("activeFocusOnClose1.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickApplicationWindow *window = helper.appWindow;
     window->show();
     window->requestActivate();
@@ -605,6 +607,7 @@ void tst_QQuickPopup::activeFocusOnClose2()
     // calling forceActiveFocus() on another item) before it closes doesn't
     // affect the active focus item when it closes.
     QQuickApplicationHelper helper(this, QStringLiteral("activeFocusOnClose2.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickApplicationWindow *window = helper.appWindow;
     window->show();
     window->requestActivate();
@@ -641,6 +644,7 @@ void tst_QQuickPopup::activeFocusOnClose3()
     // Test that a closing popup that had focus doesn't steal focus from
     // another popup that the focus was transferred to.
     QQuickApplicationHelper helper(this, QStringLiteral("activeFocusOnClose3.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickApplicationWindow *window = helper.appWindow;
     window->show();
     window->requestActivate();
@@ -672,6 +676,7 @@ void tst_QQuickPopup::activeFocusOnClosingSeveralPopups()
 {
     // Test that active focus isn't lost when multiple popup closing simultaneously
     QQuickApplicationHelper helper(this, QStringLiteral("activeFocusOnClosingSeveralPopups.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickApplicationWindow *window = helper.appWindow;
     window->show();
     window->requestActivate();
@@ -733,6 +738,7 @@ void tst_QQuickPopup::hover()
     QFETCH(bool, modal);
 
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickWindow *window = helper.window;
     window->show();
     window->requestActivate();
@@ -794,7 +800,9 @@ void tst_QQuickPopup::wheel_data()
 static bool sendWheelEvent(QQuickItem *item, const QPoint &localPos, int degrees)
 {
     QQuickWindow *window = item->window();
-    QWheelEvent wheelEvent(localPos, item->window()->mapToGlobal(localPos), QPoint(0, 0), QPoint(0, 8 * degrees), 0, Qt::Vertical, Qt::NoButton, 0);
+    QWheelEvent wheelEvent(localPos, item->window()->mapToGlobal(localPos), QPoint(0, 0),
+                           QPoint(0, 8 * degrees), Qt::NoButton, Qt::NoModifier, Qt::NoScrollPhase,
+                           false);
     QSpontaneKeyEvent::setSpontaneous(&wheelEvent);
     return qGuiApp->notify(window, &wheelEvent);
 }
@@ -805,6 +813,7 @@ void tst_QQuickPopup::wheel()
     QFETCH(bool, modal);
 
     QQuickApplicationHelper helper(this, source);
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickWindow *window = helper.window;
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window));
@@ -875,6 +884,7 @@ void tst_QQuickPopup::parentDestroyed()
 void tst_QQuickPopup::nested()
 {
     QQuickApplicationHelper helper(this, QStringLiteral("nested.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickWindow *window = helper.window;
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window));
@@ -902,6 +912,7 @@ void tst_QQuickPopup::nested()
 void tst_QQuickPopup::grabber()
 {
     QQuickApplicationHelper helper(this, QStringLiteral("grabber.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickWindow *window = helper.window;
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window));
@@ -954,6 +965,7 @@ void tst_QQuickPopup::cursorShape()
     // Ensure that the mouse cursor has the correct shape when over a popup
     // which is itself over an item with a different shape.
     QQuickApplicationHelper helper(this, QStringLiteral("cursor.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickApplicationWindow *window = helper.appWindow;
     centerOnScreen(window);
     moveMouseAway(window);
@@ -972,6 +984,7 @@ void tst_QQuickPopup::cursorShape()
 
     // Move the mouse over the text field.
     const QPoint textFieldPos(popup->x() - 10, textField->height() / 2);
+    QVERIFY(textField->contains(textField->mapFromScene(textFieldPos)));
     QTest::mouseMove(window, textFieldPos);
     QCOMPARE(window->cursor().shape(), textField->cursor().shape());
 
@@ -996,7 +1009,7 @@ void tst_QQuickPopup::componentComplete()
 
     QQmlEngine engine;
     QQmlComponent component(&engine);
-    component.setData("import QtQuick.Controls 2.2; Popup { }", QUrl());
+    component.setData("import QtQuick.Controls; Popup { }", QUrl());
 
     FriendlyPopup *qmlPopup = static_cast<FriendlyPopup *>(component.beginCreate(engine.rootContext()));
     QVERIFY(qmlPopup);
@@ -1011,6 +1024,7 @@ void tst_QQuickPopup::closeOnEscapeWithNestedPopups()
     // Tests the scenario in the Gallery example, where there are nested popups that should
     // close in the correct order when the Escape key is pressed.
     QQuickApplicationHelper helper(this, QStringLiteral("closeOnEscapeWithNestedPopups.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickApplicationWindow *window = helper.appWindow;
     window->show();
     QVERIFY(QTest::qWaitForWindowExposed(window));
@@ -1074,6 +1088,7 @@ void tst_QQuickPopup::closeOnEscapeWithNestedPopups()
 void tst_QQuickPopup::closeOnEscapeWithVisiblePopup()
 {
     QQuickApplicationHelper helper(this, QStringLiteral("closeOnEscapeWithVisiblePopup.qml"));
+    QVERIFY2(helper.ready, helper.failureMessage());
     QQuickWindow *window = helper.window;
     window->show();
     QVERIFY(QTest::qWaitForWindowActive(window));
@@ -1124,6 +1139,7 @@ void tst_QQuickPopup::orientation()
     QFETCH(QPointF, position);
 
     QQuickApplicationHelper helper(this, "orientation.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->reportContentOrientationChange(orientation);
@@ -1163,6 +1179,7 @@ void tst_QQuickPopup::qquickview()
 void tst_QQuickPopup::disabledPalette()
 {
     QQuickApplicationHelper helper(this, "disabledPalette.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -1171,35 +1188,35 @@ void tst_QQuickPopup::disabledPalette()
     QQuickPopup *popup = window->property("popup").value<QQuickPopup*>();
     QVERIFY(popup);
 
-    QSignalSpy popupEnabledSpy(popup, SIGNAL(enabledChanged()));
+    QSignalSpy popupEnabledSpy(popup, &QQuickPopup::enabledChanged);
     QVERIFY(popupEnabledSpy.isValid());
-    QSignalSpy popupPaletteSpy(popup, SIGNAL(paletteChanged()));
+    QSignalSpy popupPaletteSpy(popup, &QQuickPopup::paletteChanged);
     QVERIFY(popupPaletteSpy.isValid());
 
-    QSignalSpy popupItemEnabledSpy(popup->popupItem(), SIGNAL(enabledChanged()));
+    QSignalSpy popupItemEnabledSpy(popup->popupItem(), &QQuickItem::enabledChanged);
     QVERIFY(popupItemEnabledSpy.isValid());
-    QSignalSpy popupItemPaletteSpy(popup->popupItem(), SIGNAL(paletteChanged()));
+    QSignalSpy popupItemPaletteSpy(popup->popupItem(), &QQuickItem::paletteChanged);
     QVERIFY(popupItemPaletteSpy.isValid());
 
-    QPalette palette = popup->palette();
-    palette.setColor(QPalette::Active, QPalette::Base, Qt::green);
-    palette.setColor(QPalette::Disabled, QPalette::Base, Qt::red);
-    popup->setPalette(palette);
-    QCOMPARE(popupPaletteSpy.count(), 1);
-    QCOMPARE(popupItemPaletteSpy.count(), 1);
+    auto palette = QQuickPopupPrivate::get(popup)->palette();
+    palette->setBase(Qt::green);
+    palette->disabled()->setBase(Qt::red);
+    QCOMPARE(popupPaletteSpy.count(), 2);
+    QCOMPARE(popupItemPaletteSpy.count(), 2);
     QCOMPARE(popup->background()->property("color").value<QColor>(), Qt::green);
 
     popup->setEnabled(false);
     QCOMPARE(popupEnabledSpy.count(), 1);
     QCOMPARE(popupItemEnabledSpy.count(), 1);
-    QCOMPARE(popupPaletteSpy.count(), 2);
-    QCOMPARE(popupItemPaletteSpy.count(), 2);
+    QCOMPARE(popupPaletteSpy.count(), 3);
+    QCOMPARE(popupItemPaletteSpy.count(), 3);
     QCOMPARE(popup->background()->property("color").value<QColor>(), Qt::red);
 }
 
 void tst_QQuickPopup::disabledParentPalette()
 {
     QQuickApplicationHelper helper(this, "disabledPalette.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -1218,12 +1235,11 @@ void tst_QQuickPopup::disabledParentPalette()
     QSignalSpy popupItemPaletteSpy(popup->popupItem(), SIGNAL(paletteChanged()));
     QVERIFY(popupItemPaletteSpy.isValid());
 
-    QPalette palette = popup->palette();
-    palette.setColor(QPalette::Active, QPalette::Base, Qt::green);
-    palette.setColor(QPalette::Disabled, QPalette::Base, Qt::red);
-    popup->setPalette(palette);
-    QCOMPARE(popupPaletteSpy.count(), 1);
-    QCOMPARE(popupItemPaletteSpy.count(), 1);
+    auto palette = QQuickPopupPrivate::get(popup)->palette();
+    palette->setBase(Qt::green);
+    palette->disabled()->setBase(Qt::red);
+    QCOMPARE(popupPaletteSpy.count(), 2);
+    QCOMPARE(popupItemPaletteSpy.count(), 2);
     QCOMPARE(popup->background()->property("color").value<QColor>(), Qt::green);
 
     // Disable the overlay (which is QQuickPopupItem's parent) to ensure that
@@ -1236,8 +1252,8 @@ void tst_QQuickPopup::disabledParentPalette()
     QCOMPARE(popup->background()->property("color").value<QColor>(), Qt::red);
     QCOMPARE(popupEnabledSpy.count(), 1);
     QCOMPARE(popupItemEnabledSpy.count(), 1);
-    QCOMPARE(popupPaletteSpy.count(), 2);
-    QCOMPARE(popupItemPaletteSpy.count(), 2);
+    QCOMPARE(popupPaletteSpy.count(), 3);
+    QCOMPARE(popupItemPaletteSpy.count(), 3);
 
     popup->close();
     QTRY_VERIFY(!popup->isVisible());
@@ -1246,6 +1262,7 @@ void tst_QQuickPopup::disabledParentPalette()
 void tst_QQuickPopup::countChanged()
 {
     QQuickApplicationHelper helper(this, "countChanged.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
@@ -1262,15 +1279,16 @@ void tst_QQuickPopup::countChanged()
 // QTBUG-73243
 void tst_QQuickPopup::toolTipCrashOnClose()
 {
+    if (QOperatingSystemVersion::current() <= QOperatingSystemVersion::Windows7)
+        QSKIP("Test requires a version of Windows newer than 7: QTBUG-84443");
+    if (!canImportModule("import QtGraphicalEffects; DropShadow {}"))
+        QSKIP("Test requires QtGraphicalEffects");
+
     QQuickApplicationHelper helper(this, "toolTipCrashOnClose.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     window->show();
-    // The warning only occurs with debug builds for some reason.
-    // In any case, the warning is irrelevant, but using ShaderEffectSource is important, so we ignore it.
-#ifdef QT_DEBUG
-    QTest::ignoreMessage(QtWarningMsg, "ShaderEffectSource: 'recursive' must be set to true when rendering recursively.");
-#endif
     QVERIFY(QTest::qWaitForWindowActive(window));
 
     QTest::mouseMove(window, QPoint(window->width() / 2, window->height() / 2));
@@ -1282,13 +1300,18 @@ void tst_QQuickPopup::toolTipCrashOnClose()
 
 void tst_QQuickPopup::setOverlayParentToNull()
 {
+    if (QOperatingSystemVersion::current() <= QOperatingSystemVersion::Windows7)
+        QSKIP("Test requires a version of Windows newer than 7: QTBUG-84443");
+    if (!canImportModule("import QtGraphicalEffects; DropShadow {}"))
+        QSKIP("Test requires QtGraphicalEffects");
+
     QQuickApplicationHelper helper(this, "toolTipCrashOnClose.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
+    centerOnScreen(window);
+    moveMouseAway(window);
     window->show();
-#ifdef QT_DEBUG
-    QTest::ignoreMessage(QtWarningMsg, "ShaderEffectSource: 'recursive' must be set to true when rendering recursively.");
-#endif
     QVERIFY(QTest::qWaitForWindowActive(window));
 
     QVERIFY(QMetaObject::invokeMethod(window, "nullifyOverlayParent"));
@@ -1300,9 +1323,67 @@ void tst_QQuickPopup::setOverlayParentToNull()
     // While nullifying the overlay parent doesn't make much sense, it shouldn't crash.
 }
 
+void tst_QQuickPopup::tabFence()
+{
+    if (QGuiApplication::styleHints()->tabFocusBehavior() != Qt::TabFocusAllControls)
+        QSKIP("This platform only allows tab focus for text controls");
+
+    QQuickApplicationHelper helper(this, "tabFence.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QQuickPopup *popup = window->property("dialog").value<QQuickPopup*>();
+    QVERIFY(popup);
+    popup->open();
+    popup->setModal(true);
+
+    QQuickButton *outsideButton1 = window->property("outsideButton1").value<QQuickButton*>();
+    QVERIFY(outsideButton1);
+    QQuickButton *outsideButton2 = window->property("outsideButton2").value<QQuickButton*>();
+    QVERIFY(outsideButton2);
+    QQuickButton *dialogButton1 = window->property("dialogButton1").value<QQuickButton*>();
+    QVERIFY(dialogButton1);
+    QQuickButton *dialogButton2 = window->property("dialogButton2").value<QQuickButton*>();
+    QVERIFY(dialogButton2);
+
+    // When modal, focus loops between the two external buttons
+    outsideButton1->forceActiveFocus();
+    QVERIFY(outsideButton1->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(outsideButton2->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(outsideButton1->hasActiveFocus());
+
+    // Same thing for dialog's buttons
+    dialogButton1->forceActiveFocus();
+    QVERIFY(dialogButton1->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(dialogButton2->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(dialogButton1->hasActiveFocus());
+
+    popup->setModal(false);
+
+    // When not modal, focus goes in and out of the dialog
+    outsideButton1->forceActiveFocus();
+    QVERIFY(outsideButton1->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(outsideButton2->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(dialogButton1->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(dialogButton2->hasActiveFocus());
+    QTest::keyClick(window, Qt::Key_Tab);
+    QVERIFY(outsideButton1->hasActiveFocus());
+}
+
 void tst_QQuickPopup::invisibleToolTipOpen()
 {
     QQuickApplicationHelper helper(this, "invisibleToolTipOpen.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
 
     QQuickWindow *window = helper.window;
     centerOnScreen(window);
@@ -1315,6 +1396,9 @@ void tst_QQuickPopup::invisibleToolTipOpen()
     QObject *loader = qvariant_cast<QObject *>(window->property("loader"));
     QVERIFY(loader);
 
+    // Send an extra move event, otherwise the test fails on subsequent runs for different styles for some reason...
+    // As an added bonus, this is also slightly more realistic. :D
+    QTest::mouseMove(window, QPoint(mouseArea->width() / 2 - 1, mouseArea->height() / 2 - 1));
     QTest::mouseMove(window, QPoint(mouseArea->width() / 2, mouseArea->height() / 2));
     QTRY_VERIFY(mouseArea->property("isToolTipVisible").toBool());
 
@@ -1327,6 +1411,51 @@ void tst_QQuickPopup::invisibleToolTipOpen()
     QTRY_VERIFY(mouseArea->property("isToolTipVisible").toBool());
 }
 
+void tst_QQuickPopup::centerInOverlayWithinStackViewItem()
+{
+    QQuickApplicationHelper helper(this, "centerInOverlayWithinStackViewItem.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowExposed(window));
+
+    QQuickPopup *popup = window->property("popup").value<QQuickPopup*>();
+    QVERIFY(popup);
+    QTRY_COMPARE(popup->isVisible(), true);
+
+    // Shouldn't crash on exit.
+}
+
+void tst_QQuickPopup::destroyDuringExitTransition()
+{
+    QQuickApplicationHelper helper(this, "destroyDuringExitTransition.qml");
+    QVERIFY2(helper.ready, helper.failureMessage());
+
+    QQuickWindow *window = helper.window;
+    window->show();
+    QVERIFY(QTest::qWaitForWindowActive(window));
+
+    QPointer<QQuickPopup> dialog2 = window->property("dialog2").value<QQuickPopup*>();
+    QVERIFY(dialog2);
+    QTRY_COMPARE(dialog2->isVisible(), true);
+
+    // Close the second dialog, destroying it before its exit transition can finish.
+    QTest::keyClick(window, Qt::Key_Escape);
+    QTRY_VERIFY(!dialog2);
+
+    // Events should go through to the dialog underneath.
+    QQuickPopup *dialog1 = window->property("dialog1").value<QQuickPopup*>();
+    QVERIFY(dialog1);
+    QQuickButton *button = dialog1->property("button").value<QQuickButton*>();
+    QVERIFY(button);
+    const auto buttonClickPos = button->mapToScene(QPointF(button->width() / 2, button->height() / 2)).toPoint();
+    QTest::mousePress(window, Qt::LeftButton, Qt::NoModifier, buttonClickPos);
+    QVERIFY(button->isDown());
+    QTest::mouseRelease(window, Qt::LeftButton, Qt::NoModifier, buttonClickPos);
+    QVERIFY(!button->isDown());
+}
+
 QTEST_QUICKCONTROLS_MAIN(tst_QQuickPopup)
 
 #include "tst_qquickpopup.moc"
